with Ada.Integer_Text_IO, Globals, Strings, Integer_Sqrt;
use Globals, Strings;

pragma Elaborate_All (Ada.Integer_Text_IO,Globals,Strings,Integer_Sqrt);

package body Vectors is

  package Int_IO renames Ada.Integer_Text_IO;

  function EffFirst(V: Vector) return Integer is
  begin
    for N in V'First .. V'Last-1 loop
      if not IsZero(V(N)) then return N; end if;
    end loop;
    return V'Last;
  end EffFirst;

  function EffLast(V: Vector) return Integer is
  begin
    for N in reverse V'First+1 .. V'Last loop
      if not IsZero(V(N)) then return N; end if;
    end loop;
    return V'First;
  end EffLast;

  procedure Trim(F,L: in Integer; V: in out Vector) is
  begin
    for K in V'Range loop
      if (K<F) or (K>L) then SetZero(V(K)); end if;
    end loop;
  end Trim;

  -----------------------

  function "="(V1,V2: Vector) return Boolean is
  begin
    if (V1'First /= V2'First) or else (V1'Last /= V2'Last) then return False; end if;
    for N in V1'Range loop
      if not (V1(N)=V2(N)) then return False; end if;
    end loop;
    return True;
  end "=";

  procedure SetZero(F,L: in Integer; V: in out Vector) is
  begin
    for N in F .. L loop
      SetZero(V(N));
    end loop;
  end SetZero;

  function ZeroVector(F,L: Integer) return Vector is
    V: Vector(F..L);
  begin
    SetZero(V);
    return V;
  end ZeroVector;

  ------------------

  procedure Show1(N: in String; V: in Vector; Hide0: in Boolean := True) is
    W: constant Positive := 4;
    AllZero: Boolean := True;
  begin
    for K in V'Range loop
      if not IsZero(V(K)) then
        Show1(N & Strng(K,W) & " ",V(K));
        AllZero := False;
      end if;
    end loop;
    if not Hide0 then
      if AllZero then Show0(N & "0"); end if;
      Show0;
    end if;
  end Show1;

  procedure Show2(N: in String; V1,V2: in Vector; Hide0: in Boolean := True) is
    W: constant Positive := 4;
    KFirst: constant Integer := Integer'Min(V1'First,V2'First);
    KLast: constant Integer := Integer'Max(V1'Last,V2'Last);
    AllZero: Boolean := True;
  begin
    for K in KFirst .. KLast loop
      if not (IsZero(V1(K)) and then IsZero(V2(K))) then
        Show2(N & Strng(K,W) & " ",V1(K),V2(K));
        AllZero := False;
      end if;
    end loop;
    if not Hide0 then
      if AllZero then Show0(N & "0 0"); end if;
      Show0;
    end if;
  end Show2;

  procedure PutDim(F: in File_Type; L: in Positive) is
  begin
    Int_IO.Put(F,L);
    New_Line(F);
  end PutDim;

  procedure PutCoeffs(F: in File_Type; V: in Vector; Decimal: in Boolean := False) is
  begin
    for I in V'Range loop
      Put(F,V(I),Decimal);
    end loop;
  end PutCoeffs;

  procedure Put(F: in File_Type; V: in Vector; Decimal: in Boolean := False) is
  begin
    PutDim(F,V'Length);
    PutCoeffs(F,V,Decimal);
  end Put;

  function Get(F: File_Type; First: Integer; Decimal: Boolean := False) return Vector is
    L: Integer;
  begin
    Int_IO.Get(F,L);
    declare
      subtype VRange is Integer range First .. First+L-1;
      V: Vector(VRange);
    begin
      for M in VRange loop
        Get(F,V(M),Decimal);
      end loop;
      return V;
    end;
  end Get;

  procedure Get(F: in File_Type; V: in out Vector; Decimal: in Boolean := False) is
    Trunc: Boolean := False;
    L: Integer;
    C: Component;
  begin
    SetZero(V);
    Int_IO.Get(F,L);
    declare
      subtype VRange is Integer range V'First .. V'First+L-1;
    begin
      for M in VRange loop
        if M <= V'Last then
          Get(F,V(M),Decimal);
        else
          Get(F,C,Decimal);
          Trunc := Trunc or else not IsZero(C);
        end if;
      end loop;
    end;
    if Trunc then
      Show0("Vectors.Get Warning: input vector truncated");
    end if;
  end Get;

  procedure Write(FileName: in String; V: in Vector; Decimal: in Boolean := False) is
    F: File_Type;
  begin
    if Verbosity>0 then Show0("Writing " & FileName); end if;
    Create(F,Out_File,FileName);
    Put(F,V,Decimal);
    Close(F);
  end Write;

  function Read(FileName: String; First: Integer; Decimal: Boolean := False) return Vector is
    F: File_Type;
  begin
    if Verbosity>0 then Show0("Reading " & FileName); end if;
    Open(F,In_File,FileName);
    declare
      V: constant Vector := Get(F,First,Decimal);
    begin
      Close(F);
      return V;
    end;
  end Read;

  procedure Read(FileName: in String; V: in out Vector; Decimal: in Boolean := False) is
    F: File_Type;
  begin
    if Verbosity>0 then Show0("Reading " & FileName); end if;
    Open(F,In_File,FileName);
    Get(F,V,Decimal);
    Close(F);
  end Read;

  procedure Append(FileName: in String; V: in Vector; Decimal: in Boolean := False) is
    F: File_Type;
  begin
    if Verbosity>0 then Show0("Appending " & FileName); end if;
    begin
      Open(F,Append_File,FileName);
    exception
      when Name_Error => Create(F,Out_File,FileName);
    end;
    Put(F,V,Decimal);
    Close(F);
  end Append;

  --- Polynom1 --------------------------------------------------

  function QuasiDeg(P: Polynom1) return Natural is
  begin
    for D in reverse 1 .. P'Last loop
      if not IsZero(P(D)) then return D; end if;
    end loop;
    return 0;
  end QuasiDeg;

  function QuasiDeg0(P: Polynom1) return Boolean is
  begin
    for D in reverse 1 .. P'Last loop
      if not IsZero(P(D)) then return False; end if;
    end loop;
    return True;
  end QuasiDeg0;

  function EffDeg(P: Polynom1) return Natural is
  begin
    for D in reverse 1 .. P'Last loop
      if not IsZero(P(D)) then return D; end if;
    end loop;
    return 0;
  end EffDeg;

  procedure SetCoeff(I: in Natural; C: in Component; P: in out Polynom1) is
  begin
    Copy(C,P(I));
  end SetCoeff;

  procedure Coefficient(P: in Polynom1; I: in Natural; C: in out Component) is
  begin
    if I>P'Last then
      SetZero(C);
    else
      Copy(P(I),C);
    end if;
  end Coefficient;

  --- Polynom2 ---------------------------------------------------------

  -- index = (D*(D+1))/2+J = (D*(D+3))/2-I
  -- where D=I+J
  -- for max degree D need Polynom2(0 .. D*(D+3)/2)
  -- enumeration is
  --  J
  --  |
  --  9
  --  5 8
  --  2 4 7
  --  0 1 3 6 -- I

  function MaxDeg(P: Polynom2) return Natural is
    K: constant Natural := P'Last;
    D: constant Integer := Integer_Sqrt(2*K);
  begin
    if ((D*(D+1))/2) <= K then
      return D;
    else
      return D-1;
    end if;
  end MaxDeg;

  function EffDeg(P: Polynom2) return Natural is
    D: Integer;
  begin
    for K in reverse 1 .. P'Last loop
      if not IsZero(P(K)) then
        D := Integer_Sqrt(2*K);
        if ((D*(D+1))/2) <= K then
          return D;
        else
          return D-1;
        end if;
      end if;
    end loop;
    return 0;
  end EffDeg;

  procedure SetCoeff(I,J: in Natural; C: in Component; P: in out Polynom2) is
    D: constant Natural := I+J;
  begin
    Copy(C,P((D*(D+1))/2+J));
  end SetCoeff;

  procedure Coefficient(P: in Polynom2; I,J: in Natural; C: in out Component) is
    D: constant Natural := I+J;
  begin
    Copy(P(((D*(D+1))/2+J)),C);
  end Coefficient;

  procedure Show1(N: in String; P: in Polynom2; Hide0: in Boolean := True) is
    W: constant Positive := 4;
    Deg: constant Natural := EffDeg(P);
    AllZero: Boolean := True;
    J,K: Integer;
  begin
    for D in 0 .. Deg loop
      for I in 0 .. D loop
        J := D-I;
        K := (D*(D+1))/2+J;
        if not IsZero(P(K)) then
          Show1(N & Strng(I,W) & Strng(J,W) & " ",P(K));
          AllZero := False;
        end if;
      end loop;
    end loop;
    if not Hide0 then
      if AllZero then Show0(N & "0"); end if;
      Show0;
    end if;
  end Show1;

  procedure Show2(N: in String; P1,P2: in Polynom2; Hide0: in Boolean := True) is
    W: constant Positive := 4;
    Deg: constant Natural := Natural'Max(EffDeg(P1),EffDeg(P2));
    AllZero: Boolean := True;
    J,K: Integer;
  begin
    for D in 0 .. Deg loop
      for I in 0 .. D loop
        J := D-I;
        K := (D*(D+1))/2+J;
        if not (IsZero(P1(K)) and then IsZero(P2(K))) then
          Show2(N & Strng(I,W) & Strng(J,W) & " ",P1(K),P2(K));
          AllZero := False;
        end if;
      end loop;
    end loop;
    if not Hide0 then
      if AllZero then Show0(N & "0 0"); end if;
      Show0;
    end if;
  end Show2;

end Vectors;
