with Globals, Strings;
use Globals, Strings;
with ScalVectors.RNum;

pragma Elaborate_All (Globals,Strings);

package body Zernikes.GFix is

  use Lin,Lin_Contr;

  procedure CheckSymm(F: in Zernike; Symm: in Natural) is
    G: Zernike;
  begin
    if Symm=1 then
      RotatePi(F,G);
      Neg(G);
      if G /= F then raise Sorry with "F: Pi-symmetry check failed"; end if;
    elsif Symm=2 then
      RotatePiHalf(F,G);
      Neg(G);
      if G /= F then raise Sorry with "F: Pi/2-symmetry check failed"; end if;
    elsif Symm=0 then
      null;
    else
      raise Undefined;
    end if;
  end CheckSymm;

  procedure CheckNoSymm(F: in Zernike; Symm: in Natural) is
    G: Zernike;

    function Zero_G(M,I: Natural) return Logical is -- inefficient
      Q: constant ZModes := Make(M+2*I,Rho,Par,False,False);
      S: Scalar;
    begin
      for J in Q'Range loop
        Extract(Q(J),G,S,False);
        if (not Q(J).Err) and then Q(J).M=M and then Q(J).I=I then return Contains0(S); end if;
      end loop;
      return Uncertain;
    end Zero_G;

  begin
    if Symm=0 then
      Prod(F,F,G);
      if Zero_G(1,0)/=False then raise Sorry with "F: no-symmetry check failed"; end if;
    elsif Symm=1 then
      Copy(F,G);
      if Zero_G(1,0)/=False then raise Sorry with "F: non-radial check failed"; end if;
    elsif Symm=2 then
      Copy(F,G);
      if Zero_G(2,0)/=False then raise Sorry with "F: non-radial check failed"; end if;
    else
      raise Undefined;
    end if;
  end CheckNoSymm;

  procedure CheckSymm(M: in SMatrix; Symm: in Natural) is
    --- also checks invertibility of M-I
    A: QC_Matrix;
  begin
    if Symm=1 then
      declare
        procedure Enclose is new Enclosure (LinOp => RotatePi);
      begin
        Enclose(QC,A);
        if not Commute4Sure(M,A) then raise Sorry with "M: Pi-symmetry check failed"; end if;
      end;
    elsif Symm=2 then
      declare
        procedure Enclose is new Enclosure (LinOp => RotatePiHalf);
      begin
        Enclose(QC,A);
        if not Commute4Sure(M,A) then raise Sorry with "M: Pi/2-symmetry check failed"; end if;
      end;
    end if;
    Copy(M,A);
    AddConst(-1,A);
    Invert(A);
  end CheckSymm;

  procedure MakeWF2(P: in Model; F: in Zernike; WF,WFF: in out Zernike) is
    W: Zernike renames WFF;
  begin
    SetZero(F.R,F.P,W);
    if P.F>0 then
      AddPower(SOne,P.F,W);
      Prod(F,W,WF);
    elsif P.Hill=Zero and then P.Ring=Zero then -- W=1
      Copy(F,WF);
    else
      declare
        Hill: constant Scalar := Scal(P.Hill);
        Ring: constant Scalar := Scal(P.Ring);
      begin
        AddHill(Hill,W);
        AddRing(Ring,W);
        Add(SOne-Hill-Ring,W);
      end;
      Prod(F,W,WF);
    end if;
    Prod(WF,F,WFF);
  end MakeWF2;

  procedure MakeWF2(P: in Model; F: in Zernike; WFF: in out Zernike) is
    WF: Zernike;
  begin
    MakeWF2(P,F,WF,WFF);
  end MakeWF2;

  function MakeWF2(P: Model; F: Zernike) return Zernike is
    WF,WFF: Zernike;
  begin
    MakeWF2(P,F,WF,WFF);
    return WFF;
  end MakeWF2;

  procedure GMap0(P: in Model; F: in Zernike; WFFF,GF: in out Zernike) is
    WFF: Zernike renames GF;
  begin
    MakeWF2(P,F,WFF);
    Prod(WFF,F,WFFF);
    InvNegLap(WFFF,GF);
  end GMap0;

  procedure GMap(P: in Model; F: in Zernike; GF: in out Zernike) is
    WFF: Zernike renames GF;
    WFFF: Zernike;
  begin
    MakeWF2(P,F,WFF);
    Prod(WFF,F,WFFF);
    InvNegLap(WFFF,GF);
  end GMap;

  procedure DGMap0(WFF,H: in Zernike; WFFH,DGH: in out Zernike) is
  begin
    Prod(WFF,H,WFFH);
    InvNegLap(WFFH,DGH);
    Mult(SThree,DGH);
    if Not_STrunc and then (DGH.R /= Rho) then raise Domain_Error; end if;
  end DGMap0;

  procedure DGMap(WFF,H: in Zernike; DGH: in out Zernike) is
    WFFH: Zernike;
  begin
    Prod(WFF,H,WFFH);
    InvNegLap(WFFH,DGH);
    Mult(SThree,DGH);
    if Not_STrunc and then (DGH.R /= Rho) then raise Domain_Error; end if;
  end DGMap;

  procedure DerMatrix(P: in Model; F: in Zernike; A: in out SMatrix) is
    WFF: constant Zernike := MakeWF2(P,F);
    procedure DG(H: in Zernike; DGH: in out Zernike) is begin DGMap(WFF,H,DGH); end DG;
    procedure Enclose is new Enclosure (LinOp => DG);
  begin
    if not STrunc then
      raise Sorry with "DerMatrix: run me in numeric mode";
    end if;
    Enclose(QC,A);
  end DerMatrix;

  procedure ContrMatrix(P: in Model; F: in Zernike; M: in out SMatrix) is
    WFF: constant Zernike := MakeWF2(P,F);
    procedure DG(H: in Zernike; DGH: in out Zernike) is begin DGMap(WFF,H,DGH); end DG;
    procedure Enclose is new Enclosure (LinOp => DG);
  begin
    if not STrunc then
      raise Sorry with "ContrMatrix: run me in numeric mode";
    end if;
    if Verbosity>1 then Show0("running ContrMatrix"); end if;
    Enclose(QC,M);
    Contr_Matrix(M);
  end ContrMatrix;

  procedure DContrNorm(P: in Model; M: in SMatrix; F: in Zernike; PackSize: in Positive; S: out Radius) is
    Q: constant ZModes := Make(PackSize,Rho,Par,True,False); --- Simple but with Err modes
    WFF: constant Zernike := MakeWF2(P,F);
    procedure DG(H: in Zernike; DGH: in out Zernike) is begin DGMap(WFF,H,DGH); end DG;
    procedure DContract is new DContr (DMap => DG);
    procedure DC(H: in Zernike; DCH: in out Zernike) is begin DContract(QC,M,H,DCH); end DC;
    procedure DCNorm is new Op_Norm (LinOp => DC);
  begin
    if Verbosity>1 then
      Show1("DContrNorm: Q'Last = ",Q'Last);
    end if;
    DCNorm(Q,S);
  end DContrNorm;

  procedure ContrFix(P: in Model; M: in SMatrix; PackSize: in Positive; F: in out Zernike) is
    KMax: constant Radius := Flt(3)/Flt(4);
    Tiny: constant Radius := 1.0E-999;
    ANorm,Eps,Del,Rad,K: Scalar;
    F0: Zernike;
    K1: Radius;
  begin
    if Verbosity>1 then Show0("running ContrFix"); end if;
    NumTrunc(Size,F,F0);
    F0.R := Rho;
    GMap(P,F0,F);
    Sub(F0,F);
    Norm(F,Eps);
    Show1("Eps ",Sup(Eps));
    OpNOrm(QC,M,ANorm);
    Add(SOne,ANorm);
    ---
    Assign(KMax,K);
    Del := Eps/(SOne-K)+Scal(Tiny);
    Rad := ANorm*Del;
    Copy(F0,F);
    AddBall(Rad,F);
    DContrNorm(P,M,F,PackSize,K1);
    Show1("Lipschitz K ",K1);
    if K1>KMax then raise Sorry with "ContrFix failed"; end if;
    ---
    Del := Eps/(SOne-Scal(K1))+Scal(Tiny);
    Rad := ANorm*Del;
    Show1("ball Rad ",Sup(Rad));
    Copy(F0,F);
    AddBall(Rad,F);
  end ContrFix;

  --------------------------------------------
  --------------------------------------------

  procedure DGData(P: in Model; F: in Zernike; T: in out SSimDiag) is
    --- approx eigenvalues and eigenvectors of DGMap
    package SSV_RNum is new SSV.RNum;
    use SSV_RNum;
    Wr: Diag renames T.D;
    Vr: Matrix renames T.L;
    Wi: SVector(Wr'Range);
    A,Vi: SMatrix(Wr'Range,Wr'Range);
  begin
    if Not_STrunc then raise Sorry with "only for numeric mode"; end if;
    DerMatrix(P,F,A);
    Eigen(A,SVector(Wr),Wi,Vr,Vi);
  end DGData;

  procedure ApproxFix(P: in Model; F: in out Zernike; M: in out SMatrix; Eps: in Radius := Zero) is
    --- find approximate fixed point
    procedure Map(F1: in Zernike; F2: in out Zernike) is begin GMap(P,F1,F2); end Map;
    procedure Contract0 is new Contr0 (Map => Map);
    Iter: constant Integer := 1024;
    MaxCredit: constant Integer := 4;
    Credit: Integer := MaxCredit;
    Err,MinErr: Radius := 1.0E+9;
    G: Zernike;
  begin
    if Not_STrunc then raise Sorry with "only for numeric mode"; end if;
    if Verbosity>1 then Show0("running ApproxFix"); end if;
    Copy(F,G);
    for I in 1 .. Iter loop
      ContrMatrix(P,G,M);
      Contract0(QC,M,G,Err);
      Show1("ApproxFix: step " & Strng(I) & ", Err ",Err);
      if Err<MinErr then
        Copy(G,F);
        if (Err<Eps) or (Err=MinErr) then return; end if;
        MinErr := Err;
        Credit := MaxCredit;
      else
        exit when Credit=0;
        Credit := Credit-1;
      end if;
    end loop;
    Show1("ApproxFix: Err ",Err);
  end ApproxFix;

  procedure IterContr(P: in Model; F: in out Zernike; M: in out SMatrix) is
    --- improve approximate fixed point
    procedure Map(F1: in Zernike; F2: in out Zernike) is begin GMap(P,F1,F2); end Map;
    procedure Contract0 is new Contr0 (Map => Map);
    MaxCredit: constant Integer := 16;
    Credit: Integer := MaxCredit;
    Err,MinErr: Radius := One;
    G: Zernike;
  begin
    if Not_STrunc then raise Sorry with "only for numeric mode"; end if;
    if Verbosity>1 then Show0("running IterContr"); end if;
    Copy(F,G);
    ContrMatrix(P,G,M);
    while Credit>0 loop
      Contract0(QC,M,G,Err);
      Show1("IterContr: Err ",Err);
      if Err<MinErr then
        Copy(G,F);
        MinErr := Err;
        Credit := MaxCredit;
        ContrMatrix(P,G,M);
      else
        Credit := Credit-1;
      end if;
    end loop;
    Show1("IterContr: Err ",MinErr);
  end IterContr;

  --------------------------------------------
  --------------------------------------------

  procedure IProd(U,V: in H1Pair; S: in out Scalar) is
  begin
    L2Prod(U.XH,V.H,S);
  end IProd;

  procedure Copy(U: in H1Pair; V: in out H1Pair) is
  begin
    Copy(U.H,V.H);
    Copy(U.XH,V.XH);
  end Copy;

  procedure AddProd(S: in Scalar; U: in H1Pair; V: in out H1Pair) is
  begin
    AddProd(S,U.H,V.H);
    AddProd(S,U.XH,V.XH);
  end AddProd;

  procedure Complete(Siz: in Positive; U: in out H1Pairs) is
  begin
    for K in U'Range loop
      declare
        UK: H1Pair renames U(K);
      begin
        Copy(UK.XH,UK.H);
        NumTrunc(Siz,UK.H,UK.XH);
        InvNegLap(UK.XH,UK.H);
      end;
    end loop;
  end Complete;

  procedure Gram_Schmidt(V: in H1Pairs; U: in out H1Pairs) is
    F: constant Integer := V'First;
    L: constant Integer := V'Last;
    S,T: Scalar;
  begin
    for K in F .. L loop
      Copy(V(K),U(K));
    end loop;
    for J in F .. L-1 loop
      declare
        UJ: H1Pair renames U(J);
      begin
        IProd(UJ,UJ,S);
        Inv(S,T);
        Neg(T,S);               -- -1/<u_j,u_j>
        for K in J+1 .. L loop
          IProd(V(K),UJ,T);
          Mult(S,T);
          AddProd(T,UJ,U(K));   -- subtract from u_k the proj of v_k onto u_j
        end loop;
      end;
    end loop;
  end Gram_Schmidt;

  procedure Normalize(U: in out H1Pairs) is
    S,T: Scalar;
  begin
    for K in U'Range loop
      declare
        UK: H1Pair renames U(K);
      begin
        IProd(UK,UK,S);
        Inv(S,T);
        Sqrt(T,S);
        Mult(S,UK.H);
        Mult(S,UK.XH);
      end;
    end loop;
  end Normalize;

  procedure Bubble_Sort(V: in out FltVec) is
    --- non-increasing order
    Swapped: Boolean := True;
    R: Flt;
  begin
    while Swapped loop
      Swapped := False;
      for I in V'First+1 .. V'Last loop
        if V(I-1)<V(I) then
          R := V(I-1); V(I-1) := V(I); V(I) := R;
          Swapped := True;
        end if;
      end loop;
    end loop;
  end Bubble_Sort;

  procedure Show_Lower_Bounds(A: in SMatrix) is
    --- decent only if A is close to diagonal
    F: constant Integer := A'First(1);
    L: constant Integer := A'Last(1);
    E: FltVec(F..L);
    SK,SJ: Scalar;
  begin
    for K in F .. L loop
      Copy(A(K,K),SK);
      for J in F .. L loop
        if J /= K then
          Norm(A(J,K),SJ);
          Sub(SJ,SK);
        end if;
        E(K) := Inf(SK);
      end loop;
    end loop;
    Bubble_Sort(E);
    for K in F .. L loop
      Show1("eigen(" & Strng(Par) & "," & Strng(K-F+1) & ") >=",E(K));
    end loop;
  end Show_Lower_Bounds;

  procedure KBounds(WFF: in Zernike; V: in H1Pairs) is
    A: SMatrix(V'Range,V'Range);
    W_UKH: Zernike;
    U: H1Pairs(V'Range);
  begin
    Gram_Schmidt(V,U);
    Normalize(U);
    for K in V'Range loop
      Prod(WFF,U(K).H,W_UKH);
      for J in V'Range loop
        L2Prod(U(J).H,W_UKH,A(J,K));
      end loop;
    end loop;
    Mult(SThree,A);
    Show_Lower_Bounds(A);
  end KBounds;

  procedure Partial_Write(FileName: in String; D: in EData; Decimal: in Boolean := False) is
    F: File_Type;
  begin
    if Verbosity>0 then Show0("Writing " & FileName); end if;
    Create(F,Out_File,FileName);
    ---
    Put(F,D.Last);
    for K in 1 .. D.Last loop
      Put(F,D.Lambda(K),Decimal);
      Put(F,D.U(K).XH,Decimal);
    end loop;
    ---
    Close(F);
  end Partial_Write;

  procedure Partial_Read(FileName: in String; D: in out EData; Decimal: in Boolean := False) is
    Last: Positive := 1;
    F: File_Type;
  begin
    if Verbosity>0 then Show0("Reading " & FileName); end if;
    Open(F,In_File,FileName);
    ---
    Get(F,Last);
    for K in 1 .. IMin(Last,D.Last) loop
      Get(F,D.Lambda(K),Decimal);
      Get(F,D.U(K).XH,Decimal);
    end loop;
    ---
    Close(F);
    if Last>D.Last then
      Show0("Zernikes.GFix.Partial_Read Warning: input truncated");
    end if;
  end Partial_Read;

  procedure Complete(Siz: in Positive; P: in Model; F: in Zernike; D: in out EData) is
    --- assuming D.Lambda(K) and D.U(K).XH are defined
    S: Scalar;
  begin
    MakeWF2(P,F,D.WFF);
    Complete(Siz,D.U);
    for K in 1 .. D.Last loop
      declare
        Lambda: Scalar renames D.Lambda(K);
        Fac: Scalar renames D.Fac(K);
        U: H1Pair renames D.U(K);
      begin
        L2Prod(U.XH,U.H,S);
        Quot(Lambda,S,Fac);
        Neg(Fac);
        Center(Fac);
      end;
    end loop;
  end Complete;

  procedure LLMap0(N: in Natural; D: in EData; H: in Zernike; WH,LH: in out Zernike) is
    W: Zernike renames D.WFF;
    S: Scalar;
  begin
    DGMap0(W,H,WH,LH);
    for K in 1 .. N loop       -- subtract projection terms
      declare
        Fac: Scalar renames D.Fac(K);
        U: H1Pair renames D.U(K);
      begin
        L2Prod(U.XH,H,S);
        Mult(Fac,S);
        AddProd(S,U.H,LH);
      end;
    end loop;
  end LLMap0;

  procedure LLMap(D: in EData; H: in Zernike; LH: in out Zernike) is
    WH: Zernike;
  begin
    LLMap0(D.Last,D,H,WH,LH);
  end LLmap;

  procedure LLMapSpec(D: in EData; PackSize: in Positive; S: in out Scalar) is
    procedure LL(H: in Zernike; LH: in out Zernike) is begin LLMap(D,H,LH); end LL;
    procedure LLSpec is new Spec_Radius(LinOp => LL);
    Pow2: constant Positive := 7;
    Q: constant ZModes := Make(PackSize,Rho,Par,STrunc,STrunc);
  begin
    if Verbosity>1 then
      Show1("LLMapSpec: Q'Last = ",Q'Last);
    end if;
    LLSpec(Q,S,Pow2);
  end LLMapSpec;

  -------------------------------------------
  -------------------------------------------

  procedure Project(N: in Natural; U: in H1Pairs; H: in out Zernike) is
    --- use only in numeric mode
    A,B,C: Scalar;
  begin
    for K in 1 .. N loop
      declare
        UK: H1Pair renames U(K);
      begin
        L2Prod(UK.XH,H,A);
        L2Prod(UK.XH,UK.H,B);
        Quot(A,B,C);
        Neg(C);
        AddProd(C,UK.H,H);
      end;
    end loop;
  end Project;

  procedure Trivial_Eigen(F: in Zernike; D: in out EData) is
    --- here XH stands for W*H
    W: Zernike renames D.WFF;
    Lambda: Scalar renames D.Lambda(1);
    Fac: Scalar renames D.Fac(1);
    U: H1Pair renames D.U(1);
  begin
    if Not_STrunc then raise Sorry with "only for numeric mode"; end if;
    if Par=0 then
      Copy(SThree,Lambda);
      Copy(F,U.H);
    else
      Copy(SOne,Lambda);
      DerTheta(F,U.H);
    end if;
    Prod(W,U.H,U.XH);
    Normal(Lambda,U.XH,U.H,Fac);
    Show1(" 1 Lambda =",Lambda);
  end Trivial_Eigen;

  procedure Next_Eigen(N: in Positive; D: in out EData; Eps: in Radius := Zero) is
    --- here XH stands for W*H
    Iter: constant Integer := 512;
    MaxCredit: constant Integer := 8;
    Lambda: Scalar renames D.Lambda(N);
    Fac: Scalar renames D.Fac(N);
    U: H1Pair renames D.U(N);
    Credit: Integer := MaxCredit;
    Err,MinErr: Radius := One;
    A,B: Scalar;
    H: Zernike;
  begin
    if Not_STrunc then raise Sorry with "only for numeric mode"; end if;
    if N=1 then raise Sorry with "run Trivial_Eigen"; end if;
    Simple_Random(Rho,Par,U.H);
    Project(N-1,D.U,U.H);
    for I in 1 .. Iter loop
      LLMap0(N-1,D,U.H,U.XH,H);
      L2Prod(U.XH,H,A);
      L2Prod(U.XH,U.H,B);
      Quot(A,B,Lambda);
      Swap(U.H,H);
      Mult(Inv(Sqrt(B)),U.H);
      Sub(U.H,H);
      Err := MaxNorm(H);
      if Verbosity>1 then
        Show2(Strng(N,2) & Strng(I,4) & " Lambda Err",Approx(Lambda),Err);
      end if;
      if Err <= MinErr then
        exit when (Err<Eps) or (Err=MinErr);
        MinErr := Err;
        Credit := MaxCredit;
      else
        exit when Credit=0;
        Credit := Credit-1;
      end if;
    end loop;
    Normal(Lambda,U.XH,U.H,Fac);
  end Next_Eigen;

  procedure Find_Eigen(F: in Zernike; D: in out EData; Eps: in Radius := Zero) is
    --- find eigenv 1 .. D.Last
  begin
    Trivial_Eigen(F,D);
    for K in 2 .. D.Last loop
      Next_Eigen(K,D,Eps);
    end loop;
  end Find_Eigen;

end Zernikes.GFix;
